PROJECT_ID_ENV_VAR = "MANTIK_PROJECT_ID"
DATA_REPOSITORY_ID_ENV_VAR = "MANTIK_DATA_REPOSITORY_ID"
EXPERIMENT_REPOSITORY_ID_ENV_VAR = "MANTIK_EXPERIMENT_REPOSITORY_ID"
CODE_REPOSITORY_ID_ENV_VAR = "MANTIK_CODE_REPOSITORY_ID"
MODEL_ID_ENV_VAR = "MANTIK_MODEL_ID"
CONNECTION_ID_ENV_VAR = "MANTIK_CONNECTION_ID"
RUN_ID_ENV_VAR = "MANTIK_RUN_ID"



import json
import pathlib
import typing as t

import pytest

_FILE_PATH = pathlib.Path(__file__).parent


@pytest.fixture()
def cognito_auth_response() -> t.Dict:
    path = _FILE_PATH / "../../resources/cognito/auth-response.json"
    return _open_json_file(path)


@pytest.fixture()
def cognito_refresh_response() -> t.Dict:
    path = _FILE_PATH / "../../resources/cognito/refresh-response.json"
    return _open_json_file(path)


@pytest.fixture()
def cognito_incorrect_login_credentials_response() -> t.Dict:
    path = _FILE_PATH / (
        "../../resources/cognito/" "incorrect-login-credentials-response.json"
    )
    return _open_json_file(path)


@pytest.fixture()
def cognito_user_not_found_response() -> t.Dict:
    path = _FILE_PATH / "../../resources/cognito/user-not-found-response.json"
    return _open_json_file(path)


@pytest.fixture()
def cognito_refresh_token_expired_response() -> t.Dict:
    path = (
        _FILE_PATH
        / "../../resources/cognito/refresh-token-expired-response.json"
    )
    return _open_json_file(path)


@pytest.fixture()
def cognito_refresh_token_invalid_response() -> t.Dict:
    path = (
        _FILE_PATH
        / "../../resources/cognito/refresh-token-invalid-response.json"
    )
    return _open_json_file(path)


@pytest.fixture()
def cognito_different_client_response() -> t.Dict:
    path = _FILE_PATH / "../../resources/cognito/different-client.json"
    return _open_json_file(path)


def _open_json_file(path: pathlib.Path) -> t.Dict:
    with open(path) as json_response:
        return json.load(json_response)



import uuid

import pytest


@pytest.mark.parametrize(
    ("response", "expected", "status"),
    [
        (
            {
                "allowed": True,
                "message": "This is a mock test",
                "experimentIds": None,
            },
            '"experiment_id": "0"',
            200,
        ),
        (
            {
                "allowed": False,
                "message": "This is a mock test",
                "experimentIds": None,
            },
            "Permission denied: This is a mock test",
            403,
        ),
    ],
)
def test_supported_endpoint(
    client, mock_api_permissions, response, expected, status
):
    with mock_api_permissions(
        [("POST", "mlflow/permissions/experiments", response)]
    ):
        response = client.get(
            "api/2.0/mlflow/experiments/get",
            json={
                "experiment_id": "0",
            },
            headers={"Authorization": "test-valid-token"},
        )
        assert expected in response.text
        assert response.status_code == status


def test_not_yet_supported_endpoint(client):
    response = client.get(
        "api/2.0/mlflow/runs/get",
        json={
            "run_id": "0",
        },
        headers={"Authorization": "test-valid-token"},
    )
    assert response.json["error_code"] == "RESOURCE_DOES_NOT_EXIST"
    assert response.status_code == 404


def test_request_from_api(client, api_base_url, set_unset_api_base_url):
    with set_unset_api_base_url:
        response = client.post(
            "api/2.0/mlflow/experiments/create",
            json={
                "name": uuid.uuid4().hex,
            },
            headers={
                "Authorization": "test-valid-token",
                "MantikOrigin": "MantikApi",
            },
        )
        assert response.status_code == 200, response.text



import uuid

import pytest

import mantik.compute_backend.client as _client
import mantik.compute_backend.config.core as core
import mantik.compute_backend.credentials as _credentials
import mantik.testing.token as testing_token
import mantik.utils as utils
import mantik_compute_backend.models as _models

TEST_MLFLOW_TRACKING_URI = "https://tracking.test-uri.com"
ENV_VARS = {
    _credentials.UNICORE_USERNAME_ENV_VAR: "test-user",
    _credentials.UNICORE_PASSWORD_ENV_VAR: "test-password",
    core.COMPUTE_BUDGET_ACCOUNT_ENV_VAR: "test-project",
    utils.mlflow.TRACKING_URI_ENV_VAR: TEST_MLFLOW_TRACKING_URI,
}


@pytest.fixture(autouse=True)
def set_required_env_vars(required_env_vars):
    with utils.env.env_vars_set({**required_env_vars, **ENV_VARS}):
        yield


@pytest.fixture()
@testing_token.set_token()
def compute_backend_client(
    set_required_env_vars,
) -> _client.ComputeBackendClient:
    return _client.ComputeBackendClient.from_env()


class TestComputeBackendClient:
    def test_url(self, compute_backend_client):
        expected = "https://compute.test-uri.com"

        result = compute_backend_client.url

        assert result == expected

    def test_submit_url(self, compute_backend_client):
        expected = "https://compute.test-uri.com/submit"

        result = compute_backend_client.submit_url

        assert result == expected

    def test_submit(
        self,
        requests_mock,
        compute_backend_client,
        tmp_dir_as_test_mantik_folder,
        mlproject_path,
    ):
        experiment_id = 123
        run_id = uuid.uuid4()
        unicore_job_id = "1"

        expected = _models.SubmitRunResponse(
            experiment_id=experiment_id,
            run_id=run_id,
            unicore_job_id=unicore_job_id,
        ).model_dump(mode="json")

        # Note: fastapi converts pydantic.BaseModel to dict
        # https://github.com/tiangolo/fastapi/blob/master/fastapi/routing.py#L77
        requests_mock.post(
            f"{compute_backend_client.submit_url}/{experiment_id}",
            json=expected,
            status_code=201,
        )

        result = compute_backend_client.submit_run(
            run_name="test-run-name",
            experiment_id=experiment_id,
            mlproject_path=mlproject_path,
            mlflow_parameters={"test-parameter": "test-value"},
            backend_config="compute-backend-config.json",
            entry_point="main",
        )

        assert result.status_code == 201
        assert result.json() == expected



import logging
import pathlib
import platform
import typing as t
import uuid

import click

import mantik.cli._options as main_options
import mantik.cli.runs._options as _options
import mantik.cli.runs.runs as runs
import mantik.cli.utils as utils
import mantik.runs.local as local_runs
import mantik.runs.schemas as schemas


logger = logging.getLogger(__name__)


def get_gpu_info() -> list:
    import GPUtil

    gpus = GPUtil.getGPUs()

    gpu_info_list = []

    for gpu in gpus:
        gpu_info = {
            "Name": gpu.name,
            "ID": str(gpu.id),
            "Driver": gpu.driver,
            "TotalMemory": str(gpu.memoryTotal),
        }

        gpu_info_list.append(gpu_info)

    return gpu_info_list


def get_system_details() -> dict:
    import cpuinfo
    import psutil

    system_details = {
        "Platform": platform.platform(),
        "CpuModel": cpuinfo.get_cpu_info().get("brand_raw"),
        "CpuCoreCount": psutil.cpu_count(),
        "TotalMemoryGB": psutil.virtual_memory().total / (1024**3),
        "GPUInfo": get_gpu_info(),
    }
    return {"SystemDetails": system_details}


@runs.cli.command("local")
@_options.MLPROJECT_PATH
@main_options.get_name_option(required=True, help_option="Name of the Run.")
@_options.ENTRY_POINT
@main_options.PROJECT_ID
@main_options.DATA_REPOSITORY_ID
@main_options.EXPERIMENT_REPOSITORY_ID
@main_options.CODE_REPOSITORY_ID
@_options.BRANCH
@_options.COMMIT
@click.option(
    "--backend-config-system-info",
    type=bool,
    default=True,
    help="Populate the backend config JSON with system info.",
)
@_options.PARAMETER
@main_options.VERBOSE
def run_project(
    name: str,
    mlproject_path: pathlib.Path,
    entry_point: str,
    parameter: t.List[str],
    verbose: bool,  # noqa
    project_id: uuid.UUID,
    data_repository_id: t.Optional[uuid.UUID],
    experiment_repository_id: uuid.UUID,
    code_repository_id: uuid.UUID,
    branch: t.Optional[str],
    commit: t.Optional[str],
    backend_config_system_info: bool,
) -> None:
    """Run an MLflow project locally and save the results in Mantik API

    Note that `MLPROJECT_PATH` is the relative path to the MLflow project file
    with your Code Repository as root.

    Remember that when you execute a run, the code is retrieved from your
    remote Git Code Repository. So make sure to commit and push your
    changes before executing a run!

    To find the respective required IDs make sure to check Mantik's UI

    """

    _options.check_commit_or_branch(branch=branch, commit=commit, logger=logger)

    logger.debug("Parsing MLflow entry point parameters")
    parameters = utils.dict_from_list(parameter)

    mantik_token = utils.access_token_from_env_vars()

    system_details = get_system_details() if backend_config_system_info else {}
    local_runs.run(
        data=schemas.RunConfiguration(
            name=name,
            experiment_repository_id=experiment_repository_id,
            code_repository_id=code_repository_id,
            branch=branch,
            commit=commit,
            data_repository_id=data_repository_id,
            mlflow_mlproject_file_path=mlproject_path.as_posix(),
            entry_point=entry_point,
            mlflow_parameters=parameters,
            backend_config=system_details,
        ),
        project_id=project_id,
        mantik_token=mantik_token,
    )
    click.echo("Done!")



"""Integration tests for UNICORE.

**WARNING:** The UNICORE container takes up to 15-30 seconds to process a job,
hence be careful with the number of test cases used.


"""
import pathlib

import mlflow.entities as entities
import mlflow.projects as projects
import pytest

import mantik.compute_backend.config.core as core
import mantik.compute_backend.credentials as _credentials
import mantik.utils as utils
import mantik_compute_backend.backend as backend
import mantik_compute_backend.utils as mantik_compute_backend_utils

FILE_PATH = pathlib.Path(__file__).parent


@pytest.fixture()
def required_unicore_env_vars(unicore_credentials) -> dict:
    return {
        _credentials.UNICORE_USERNAME_ENV_VAR: unicore_credentials["user"],
        _credentials.UNICORE_PASSWORD_ENV_VAR: unicore_credentials["password"],
        core.COMPUTE_BUDGET_ACCOUNT_ENV_VAR: "test-project",
        mantik_compute_backend_utils.mlflow.RUN_NAME_ENV_VAR: "my run name",
    }


RESOURCES = {
    "Queue": "batch",
    "Runtime": "1h",
    "Nodes": 1,
    "TotalCPUs": 1,
    "CPUsPerNode": 1,
    "MemoryPerNode": "1M",
}

EXPECTED_RESOURCE_MESSAGES = (
    "Queue=batch",
    "Runtime=3600",
    "Nodes=1",
    "TotalCPUs=1",
    "CPUsPerNode=1",
    "MemoryPerNode=1048576",
    "Project=test-project",
)


class TestUnicoreBackend:
    @pytest.mark.parametrize(
        (
            "entry_point",
            "params",
            "backend_config",
            "expected_job_status",
            "expected_log_messages",
        ),
        [
            # Test case: Apptainer with 1 env var
            #
            # * Test case will fail since no Apptainer image present in UNICORE
            #   container, thus the post command on login node won't be executed
            (
                "main",
                {"print": "test"},
                {
                    "Environment": {
                        "Variables": {"TEST_VAR": "test_value"},
                        "Apptainer": {
                            "Path": "/does/not/exist.sif",
                            "Type": "remote",
                        },
                        "PreRunCommandOnLoginNode": [
                            "echo precommand login node",
                            "echo again",
                        ],
                        "PostRunCommandOnLoginNode": [
                            "echo postcommand login node"
                        ],
                        "PreRunCommandOnComputeNode": [
                            "echo precommand compute node"
                        ],
                        "PostRunCommandOnComputeNode": [
                            "echo postcommand compute node"
                        ],
                    },
                    "Resources": RESOURCES,
                },
                entities.RunStatus.FAILED,
                [
                    *EXPECTED_RESOURCE_MESSAGES,
                    "MLFLOW_RUN_ID=",
                    "; export MLFLOW_RUN_ID",
                    'TEST_VAR="test_value"; export TEST_VAR',
                    'source $(dirname "$(realpath "$0")")/mantik.sh',
                ],
            ),
            # Test case: Python with 1 env var
            #
            # * Test case will fail since no venv present in UNICORE container,
            #   thus the post command on login node won't be executed
            (
                "main",
                {"print": "test"},
                {
                    "Environment": {
                        "Variables": {"TEST_VAR": "test_value"},
                        "Python": {
                            "Path": "/does/not/exist",
                        },
                        "PreRunCommandOnLoginNode": [
                            "echo precommand login node",
                            "echo again",
                        ],
                        "PostRunCommandOnLoginNode": [
                            "echo postcommand login node"
                        ],
                        "PreRunCommandOnComputeNode": [
                            "echo precommand compute node"
                        ],
                        "PostRunCommandOnComputeNode": [
                            "echo postcommand compute node"
                        ],
                    },
                    "Resources": RESOURCES,
                },
                entities.RunStatus.FAILED,
                [
                    *EXPECTED_RESOURCE_MESSAGES,
                    "MLFLOW_RUN_ID=",
                    "; export MLFLOW_RUN_ID",
                    'TEST_VAR="test_value"; export TEST_VAR',
                    "echo precommand login node",
                    "echo again",
                    'source $(dirname "$(realpath "$0")")/mantik.sh',
                ],
            ),
            # Test case: No execution environment and echo entry point
            #
            # * Should succeed
            # * Pre/postcommand will be executed on login node, hence no modules
            #   loaded here. Modules would make the pre command fail (no modules
            #   in UNICORE container), and the main application will not be
            #   executed, as well as the post command.
            (
                "echo",
                {"output": "test"},
                {
                    "Environment": {
                        "Variables": {"TEST_VAR": "test_value"},
                        "PreRunCommandOnLoginNode": [
                            "echo precommand login node",
                            "echo again",
                        ],
                        "PostRunCommandOnLoginNode": [
                            "echo postcommand login node"
                        ],
                    },
                    "Resources": RESOURCES,
                },
                entities.RunStatus.FINISHED,
                [
                    "Launched pre command <echo precommand login node && "
                    "echo again>\n",
                    "Executing command: echo precommand login node && "
                    "echo again",
                    "Execution on login node",
                    *EXPECTED_RESOURCE_MESSAGES,
                    'TEST_VAR="test_value"; export TEST_VAR',
                    "MLFLOW_RUN_ID=",
                    "; export MLFLOW_RUN_ID",
                    'source $(dirname "$(realpath "$0")")/mantik.sh',
                ],
            ),
        ],
    )
    def test_run(
        self,
        tmp_path,
        env_vars_set,
        unicore_api_url,
        required_unicore_env_vars,
        example_project_path,
        entry_point,
        params,
        backend_config,
        expected_job_status,
        expected_log_messages,
    ):
        backend_config = {
            **backend_config,
            "UnicoreApiUrl": unicore_api_url,
            # Following values are set by MLflow before running a project.
            projects.PROJECT_STORAGE_DIR: example_project_path.as_posix(),
        }

        expected_log_messages.extend(
            [
                # MLflow Tracking URI should be set by compute backend
                f'MLFLOW_TRACKING_URI="file://{tmp_path.as_posix()}/mlruns"; export MLFLOW_TRACKING_URI',  # noqa: E501
            ]
        )

        # Following env vars must be set for the config.
        tracking_uri = tmp_path / "mlruns"
        with env_vars_set(
            {
                **required_unicore_env_vars,
                # Point MLFLOW_TRACKING_URI to a temporary directory
                utils.mlflow.TRACKING_URI_ENV_VAR: tracking_uri.as_uri(),  # noqa: E501
            }
        ):
            result = backend.Backend().run(
                project_uri=example_project_path.as_posix(),
                entry_point=entry_point,
                params=params,
                version=None,
                backend_config=backend_config,
                tracking_uri=None,
                experiment_id=None,
            )
            # Wait for job to fail or succeed
            result.wait()
            # For the structure of the UNICORE job properties responce, see
            # `tests/resources/unicore-responses/job-property-response.json`.
            result_log_messages = "\n".join(result._job._job.properties["log"])
        for exp in expected_log_messages:
            assert exp in result_log_messages, f"Message {exp!r} not in logs"
        assert result.get_status() == expected_job_status



import io
import pathlib
import unittest
import uuid
import zipfile

import pytest

import mantik.runs.schemas as run_schemas
import mantik.utils.mantik_api.code_repository as code_api
import mantik.utils.mantik_api.experiment_repository as experiment_api


@pytest.fixture(scope="session")
def sample_experiment_repository_id():
    return uuid.uuid4()


@pytest.fixture(scope="session")
def sample_code_repository_id():
    return uuid.uuid4()


@pytest.fixture(scope="session")
def sample_project_id() -> uuid.UUID:
    return uuid.uuid4()


@pytest.fixture()
def sample_experiment_repository(sample_experiment_repository_id):
    return experiment_api.ExperimentRepository(
        experiment_repository_id=sample_experiment_repository_id,
        mlflow_experiment_id="123",
        name="Name",
        artifact_location="somewhere.com",
    )


@pytest.fixture()
def sample_code_repository(sample_code_repository_id):
    return code_api.CodeRepository(
        code_repository_id=sample_code_repository_id,
        code_repository_name="Name",
        uri="some/uri.git",
        access_token="1234",
    )


@pytest.fixture(scope="session")
def sample_run_configuration(
    sample_experiment_repository_id, sample_code_repository_id
) -> run_schemas.RunConfiguration:
    return run_schemas.RunConfiguration(
        name="Sample",
        experiment_repository_id=sample_experiment_repository_id,
        code_repository_id=sample_code_repository_id,
        branch="branch",
        commit="commit",
        data_repository_id=uuid.uuid4(),
        mlflow_mlproject_file_path="some/path/MLProject",
        entry_point="main",
        mlflow_parameters={"output": "hello world"},
        backend_config={},
    )


@pytest.fixture()
def fake_token():
    return "1234"


@pytest.fixture
def resource():
    return (
        pathlib.Path(__file__).parent.parent.parent
        / "resources/test-project/MLproject"
    )


@pytest.fixture
def mock_get_artifacts_url():
    yield unittest.mock.patch(
        "mantik.utils.mantik_api.run.get_download_artifact_url",
        return_value="https://fake/url/artifacts.zip?yada&&yada",
    )


@pytest.fixture
def zipped_file_name():
    return "test_file_name"


@pytest.fixture
def zipped_file_bytes(tmpdir, zipped_file_name, resource):
    def create_zip(input_file, output_zip):
        with zipfile.ZipFile(output_zip, "w") as zipf:
            zipf.write(input_file)
        with open(output_zip, "rb") as zip_file:
            byte_stream = io.BytesIO(zip_file.read())
        return byte_stream

    return create_zip(resource, tmpdir / zipped_file_name)


@pytest.fixture
def mock_get_url(zipped_file_bytes):
    yield unittest.mock.patch("requests.get", return_value=zipped_file_bytes)


@pytest.fixture
def mock_authentication():
    with unittest.mock.patch(
        "mantik.authentication.auth.get_valid_access_token",
        return_value="1234-token",
    ):
        yield



import pytest


def test_authenticate_api_calls_returns_200_for_non_api_path(client, api_path):
    response = client.get("/#/experiments/0")

    assert response.status_code == 200


@pytest.mark.parametrize(
    ("headers", "expected_status_code", "expected_message"),
    [
        ({}, 401, "'Authorization' header missing"),
        ({"Authorization": ""}, 401, "Empty token in 'Authorization' header"),
        ({"Authorization": "test-invalid-token"}, 401, "Invalid token"),
        ({"Authorization": "Bearer test-invalid-token"}, 401, "Invalid token"),
        (
            {"Authorization": "test-valid-token"},
            200,
            None,
        ),
        ({"Authorization": "Bearer test-valid-token"}, 200, None),
    ],
)
def test_authenticate_api_calls(
    client,
    api_path,
    headers,
    expected_status_code,
    expected_message,
    mock_api_permissions,
):
    with mock_api_permissions(
        [
            (
                "POST",
                "mlflow/permissions/experiments",
                {
                    "allowed": True,
                    "message": "This is a mock test",
                    "experimentIds": None,
                },
            )
        ]
    ):
        response = client.get(
            f"{api_path}/experiments/get",
            headers=headers,
            json={"experiment_id": 0},
        )

        assert response.status_code == expected_status_code
        if expected_message is not None:
            assert response.text == expected_message



import dataclasses
import typing as t
import uuid


@dataclasses.dataclass
class RunConfiguration:
    name: str
    experiment_repository_id: uuid.UUID
    code_repository_id: uuid.UUID
    branch: t.Optional[str]
    commit: t.Optional[str]
    data_repository_id: t.Optional[uuid.UUID]
    mlflow_mlproject_file_path: str
    entry_point: str
    mlflow_parameters: dict
    backend_config: dict
    mlflow_run_id: t.Optional[uuid.UUID] = None

    def to_post_payload(self) -> dict:
        return {
            "name": self.name,
            "experimentRepositoryId": str(self.experiment_repository_id),
            "codeRepositoryId": str(self.code_repository_id),
            "branch": self.branch,
            "commit": self.commit,
            "dataRepositoryId": str(self.data_repository_id)
            if self.data_repository_id is not None
            else None,
            "mlflowMlprojectFilePath": self.mlflow_mlproject_file_path,
            "entryPoint": self.entry_point,
            "mlflowParameters": self.mlflow_parameters,
            "backendConfig": self.backend_config,
            "mlflowRunId": str(self.mlflow_run_id),
        }



import datetime
import uuid

import pytest
import requests_mock

import mantik.remote_file_service.abstract_file_service as afs
import mantik.remote_file_service.data_client as _data_client
import mantik.utils.mantik_api.client as mantik_api


class TestDataClient:
    def test_list_directory(self, data_client, fake_unicore_fs):
        target = "test_target"
        assert data_client.list_directory(
            target
        ) == fake_unicore_fs.list_directory(target)

    def test_create_directory(self, data_client, fake_unicore_fs, project_id):
        with requests_mock.Mocker() as m:
            m.post(
                url=f"{mantik_api._DEFAULT_MANTIK_API_URL}/"
                f"projects/{str(project_id)}/data",
                json={
                    "dataRepositoryId": str(uuid.uuid4()),
                },
            )
            target = "test_target"
            mantik_name = ""
            mantik_description = ""
            assert data_client.create_directory(
                target,
                data_repository_name=mantik_name,
                data_repository_description=mantik_description,
            ) == fake_unicore_fs.create_directory(target)
            assert m.request_history[0].json() == {
                "accessToken": None,
                "dataRepositoryName": "test_target",
                "description": "This data repository with "
                "path: test_target was uploaded by FAKE-USER",
                "uri": "test_target",
            }

    def test_remove_directory(self, data_client, fake_unicore_fs, project_id):
        data_repository_id = uuid.uuid4()
        with requests_mock.Mocker() as m:
            m.delete(
                url=f"{mantik_api._DEFAULT_MANTIK_API_URL}/"
                f"projects/{str(project_id)}/data/{data_repository_id}",
                json={
                    "dataRepositoryId": str(uuid.uuid4()),
                },
            )
            target = "test_target"
            assert data_client.remove_directory(
                target, data_repository_id=data_repository_id
            ) == fake_unicore_fs.remove_directory(target)
            assert len(m.request_history) == 1

    def test_copy_directory(self, data_client, fake_unicore_fs, project_id):
        with requests_mock.Mocker() as m:
            m.post(
                url=f"{mantik_api._DEFAULT_MANTIK_API_URL}/"
                f"projects/{str(project_id)}/data",
                json={
                    "dataRepositoryId": str(uuid.uuid4()),
                },
            )
            source = "test_source"
            target = "test_target"
            mantik_name = "test-name"
            mantik_description = ""
            assert data_client.copy_directory(
                source,
                target,
                data_repository_name=mantik_name,
                data_repository_description=mantik_description,
            ) == fake_unicore_fs.copy_directory(source, target)
            assert m.request_history[0].json() == {
                "accessToken": None,
                "dataRepositoryName": "test-name",
                "description": "This data repository "
                "with path: test_target was uploaded by FAKE-USER",
                "uri": "test_target",
            }

    def test_create_file_if_not_exists(
        self, data_client, fake_unicore_fs, project_id
    ):
        with requests_mock.Mocker() as m:
            m.post(
                url=f"{mantik_api._DEFAULT_MANTIK_API_URL}/"
                f"projects/{str(project_id)}/data",
                json={
                    "dataRepositoryId": str(uuid.uuid4()),
                },
            )
            target = "test_target"
            mantik_name = "test-name"
            mantik_description = "test-description"
            assert data_client.create_file_if_not_exists(
                target,
                data_repository_name=mantik_name,
                data_repository_description=mantik_description,
            ) == fake_unicore_fs.create_file_if_not_exists(target)
            assert m.request_history[0].json() == {
                "accessToken": None,
                "dataRepositoryName": "test-name",
                "description": "test-description",
                "uri": "test_target",
            }

    def test_remove_file(self, data_client, fake_unicore_fs, project_id):
        data_repository_name = "data-repo-test-name"
        data_repository_id = "3fa85f64-5717-4562-b3fc-2c963f66afa6"
        with requests_mock.Mocker() as m:
            m.get(
                url=f"{mantik_api._DEFAULT_MANTIK_API_URL}/"
                f"projects/{str(project_id)}/data",
                json={
                    "totalRecords": 1,
                    "dataRepositories": [
                        {
                            "dataRepositoryId": "3fa85f64-5717-4562-b3fc-2c963f66afa6",  # noqa
                            "dataRepositoryName": data_repository_name,
                            "uri": "string",
                            "accessToken": "string",
                            "description": "string",
                        }
                    ],
                },
            )
            m.delete(
                url=f"{mantik_api._DEFAULT_MANTIK_API_URL}/projects/"
                f"{str(project_id)}/data/{data_repository_id}",
                json={
                    "dataRepositoryId": str(uuid.uuid4()),
                },
            )
            target = "test_target"
            assert data_client.remove_file(
                target, data_repository_name=data_repository_name
            ) == fake_unicore_fs.remove_file(target)
            assert len(m.request_history) == 2

    def test_copy_file(self, data_client, fake_unicore_fs, project_id):
        with requests_mock.Mocker() as m:
            m.post(
                url=f"{mantik_api._DEFAULT_MANTIK_API_URL}/"
                f"projects/{str(project_id)}/data",
                json={
                    "dataRepositoryId": str(uuid.uuid4()),
                },
            )
            source = "test_source"
            target = "test_target"
            mantik_name = "test-name"
            mantik_description = "test-description"
            assert data_client.copy_file(
                source,
                target,
                data_repository_name=mantik_name,
                data_repository_description=mantik_description,
            ) == fake_unicore_fs.copy_file(source, target)
            assert m.request_history[0].json() == {
                "accessToken": None,
                "dataRepositoryName": "test-name",
                "description": "test-description",
                "uri": "test_target",
            }

    def test_user(self, data_client, fake_unicore_fs):
        assert data_client.user == fake_unicore_fs.user

    def test_change_permissions(self, data_client, fake_unicore_fs):
        target = "test_target"
        new_permissions = afs.FileMeta(
            last_changed=datetime.datetime(2022, 1, 1),
            mode="xxx",
            owner=fake_unicore_fs.user,
        )
        assert data_client.change_permissions(
            target, new_permissions
        ) == fake_unicore_fs.change_permissions(target, new_permissions)

    def test_remove_file_no_reference(self, data_client, fake_unicore_fs):
        with pytest.raises(_data_client.DataClientException) as e:
            target = "test_target"
            data_client.remove_file(target)
        assert str(e.value) == (
            "Either 'data_repository_id' or "
            "'data_repository_name' must be passed as "
            "arguments, in order to delete "
            "the associated data repository in Mantik.\n"
            "If no reference is present use the FileService directly."
        )

    def test_remove_file_name_not_present(
        self, data_client, fake_unicore_fs, project_id
    ):
        with pytest.raises(_data_client.DataClientException) as e:
            data_repository_name = "not present name"
            with requests_mock.Mocker() as m:
                m.get(
                    url=f"{mantik_api._DEFAULT_MANTIK_API_URL}/"
                    f"projects/{str(project_id)}/data",
                    json={
                        "totalRecords": 1,
                        "dataRepositories": [
                            {
                                "dataRepositoryId": "3fa85f64-5717-4562-b3fc-2c963f66afa6",  # noqa
                                "dataRepositoryName": "name",
                                "uri": "string",
                                "accessToken": "string",
                                "description": "string",
                            }
                        ],
                    },
                )
                target = "test_target"
                data_client.remove_file(
                    target, data_repository_name=data_repository_name
                )
        assert str(e.value) == (
            "A data repository with name: not present name was not found.\n"
            "Make sure that the name is correct and "
            "that the repository has not been "
            "deleted already.\n"
            "To solve the problem, consider passing 'data_repository_id'."
        )

    def test_init_false_project_id(self, env_vars_set):
        with env_vars_set(
            {
                "MANTIK_PROJECT_ID": "error",
                "REMOTE_FS_TYPE": "UNICORE",
            }
        ), pytest.raises(_data_client.DataClientException) as e:
            assert _data_client.DataClient.from_env()

        assert (
            str(e.value)
            == "Badly formed hexadecimal UUID string for project ID"
        )

    def test_init_not_supported_backend(self, env_vars_set):
        with env_vars_set(
            {
                "MANTIK_PROJECT_ID": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
                "REMOTE_FS_TYPE": "NEW_BACKEND",
            }
        ), pytest.raises(_data_client.DataClientException) as e:
            assert _data_client.DataClient.from_env()

        expected = (
            "Invalid remote file system type, "
            "set REMOTE_FS_TYPEas one of this supported "
            "types ['S3', 'UNICORE']"
        )

        result = str(e.value)

        assert result == expected



import logging
import os
import pathlib
import tempfile
import typing as t
import uuid

import mantik.runs.schemas as schemas
import mantik.utils.mantik_api as mantik_api
import mantik.utils.mantik_api.code_repository as code_api
import mantik.utils.mantik_api.experiment_repository as experiment_api


logger = logging.getLogger(__name__)


class LocalRunManager:
    def __init__(self):
        pass

    @classmethod
    def clone_git_repo(
        cls, uri: str, branch: str, target_directory: str
    ) -> None:
        try:
            import git
        except ModuleNotFoundError as e:
            raise RuntimeError(
                "Please install git module: required to execute local runs"
            ) from e

        logger.debug(f"Cloning repo {uri}...")

        git.Repo.clone_from(uri, target_directory, branch=branch)

        logger.debug("Repo cloned.")

    @staticmethod
    def start_local_run(
        mlflow_experiment_id: str,
        data: schemas.RunConfiguration,
        project_id: uuid.UUID,
        mantik_token: str,
        uri: str,
    ) -> None:
        # Lazy import so that mantik can be used without mlflow
        import mlflow
        import mantik.mlflow
        import mantik.utils.mlflow

        logger.info(f"Starting mlflow run: '{data.name}' ...")
        mlflow.set_tracking_uri(
            os.getenv(
                mantik.utils.mlflow.TRACKING_URI_ENV_VAR,
                mantik.utils.mlflow.DEFAULT_TRACKING_URI,
            )
        )
        with mantik.mlflow.start_run(
            experiment_id=mlflow_experiment_id,
            run_name=data.name,
        ) as active_run:
            data.mlflow_run_id = active_run.info.run_id
            run_id = save_run_data(
                data=data, project_id=project_id, mantik_token=mantik_token
            )
            mantik_api.run.update_run_status(
                project_id=project_id,
                token=mantik_token,
                status=active_run.info.status,
                run_id=run_id,
            )
        try:
            _run = mlflow.run(
                uri=uri,
                backend="local",
                experiment_id=mlflow_experiment_id,
                run_name=data.name,
                entry_point=data.entry_point,
                parameters=data.mlflow_parameters,
                backend_config={
                    mantik.mlflow.get_local_backend_config_run_id_env_var(): active_run.info.run_id  # noqa: E501
                },
                build_image=True,
                docker_args={
                    "e": f"{mantik.utils.mlflow.TRACKING_TOKEN_ENV_VAR}={os.environ[mantik.utils.mlflow.TRACKING_TOKEN_ENV_VAR]}"  # noqa: E501
                },
            )
        except KeyboardInterrupt as e:
            logger.warning(
                "Keyboard interrupt, setting Mantik run status to KILLED"
            )
            mantik_api.run.update_run_status(
                project_id=project_id,
                token=mantik_token,
                status="KILLED",
                run_id=run_id,
            )
            raise e
        # MlflowException is raised when something goes wrong with mlflow
        # FileNotFoundError is raised when pyenv or conda are not found
        except (mlflow.exceptions.MlflowException, FileNotFoundError) as e:
            logger.warning(
                "Failed to execute run, setting Mantik run status to FAILED"
            )
            mantik_api.run.update_run_status(
                project_id=project_id,
                token=mantik_token,
                status="FAILED",
                run_id=run_id,
            )
            raise e
        logger.debug("Run finished successfully")
        mantik_api.run.update_run_status(
            project_id=project_id,
            token=mantik_token,
            status=_run.get_status(),
            run_id=run_id,
        )


def fetch_code_and_experiment(
    project_id: uuid.UUID,
    code_repository_id: uuid.UUID,
    experiment_repository_id: uuid.UUID,
    mantik_token: str,
) -> t.Tuple[code_api.CodeRepository, experiment_api.ExperimentRepository]:
    logger.debug("Fetching code and experiment from Mantik API...")
    code = code_api.get_one(
        code_repository_id=code_repository_id,
        project_id=project_id,
        token=mantik_token,
    )
    experiment = experiment_api.get_one(
        experiment_repository_id=experiment_repository_id,
        project_id=project_id,
        token=mantik_token,
    )
    logger.debug(f"Fetched {code} and {experiment}")
    return code, experiment


def save_run_data(
    data: schemas.RunConfiguration, project_id: uuid.UUID, mantik_token: str
) -> uuid.UUID:
    logger.debug(f"Saving data... {data}")

    response = mantik_api.run.save_run(
        project_id=project_id,
        run_data=data.to_post_payload(),
        token=mantik_token,
    )

    logger.debug("Results saved to Mantik API")
    return uuid.UUID(response.json()["runId"])


def run(
    data: schemas.RunConfiguration,
    project_id: uuid.UUID,
    mantik_token: str,
    run_manager: LocalRunManager = LocalRunManager(),
):
    code, experiment = fetch_code_and_experiment(
        project_id=project_id,
        code_repository_id=data.code_repository_id,
        experiment_repository_id=data.experiment_repository_id,
        mantik_token=mantik_token,
    )

    data.name = experiment_api.get_unique_run_name(
        experiment_repository_id=data.experiment_repository_id,
        project_id=project_id,
        token=mantik_token,
        run_name=data.name,
    )

    with tempfile.TemporaryDirectory() as temp_dir:
        run_manager.clone_git_repo(
            uri=code.uri,
            branch=data.commit or data.branch,
            target_directory=temp_dir,
        )

        run_manager.start_local_run(
            uri=path_directory_of_mlproject_file(
                str(
                    pathlib.Path(temp_dir).joinpath(
                        data.mlflow_mlproject_file_path
                    )
                )
            ),
            data=data,
            mlflow_experiment_id=experiment.mlflow_experiment_id,
            mantik_token=mantik_token,
            project_id=project_id,
        )


def path_directory_of_mlproject_file(mlproject_file_path: str) -> str:
    return str(pathlib.Path(mlproject_file_path).parent)
